open Pervasives

module S =
  Set.Make
    (struct
      type t = string
      let compare = compare
    end)

module M =
  Map.Make
    (struct
      type t = string
      let compare = compare
    end)

module M' =
  Map.Make
    (struct
      type t = string * string 
      let compare = compare
    end)

(* options *)
let ovwr = ref false
let merge = ref true

let domimax = 1024*1024
let rec mm (i,j) =
  if j>domimax then (i*domimax/j,domimax)
  else (i,j)
let counter = ref 0
let newvar n v =
  incr counter;
  String.concat "" [List.hd (Str.split (Str.regexp "[.]") v);".";string_of_int n;".";string_of_int !counter] 

let rec output_line f s =
  output_string f s;
  output_char f '\n'

(* addtion of fraction *)
let rec gcd a b =
  if b = 0 then a else gcd b (a mod b)
let rec lcm a b = a * b / gcd a b
let rec fradd (a,b) (c,d) =
  let e = lcm b d in
  (a*e/b+c*e/d,e)
let rec frmul (a,b) (c,d) =
  let x = a*c in
  if x=0 then (0,1) else
  let y = b*d in
  let z = gcd x y in
  (x/z,y/z)
let rec frdiv (a,b) (c,d) =
  frmul (a,b) (d,c)

(* add edge to map *)
let rec adde x y a i m =
  try
    let yai' = M.find x m in
    try
      let ai' = M.find y yai' in
      try
	let i' = M.find a ai' in
	M.add x (M.add y (M.add a (fradd i i') ai') yai') m
      with Not_found -> M.add x (M.add y (M.add a i ai') yai') m
    with Not_found -> M.add x (M.add y (M.singleton a i) yai') m
  with Not_found -> M.add x (M.singleton y (M.singleton a i)) m
(* add edges to map *)
let rec addes x y ai m =
  try
    let yai' = M.find x m in
    M.add x (M.add y ai yai') m
  with Not_found -> M.add x (M.singleton y ai) m
(* merge edges *)
let rec mergee ai ai' =
  let f a i ai =
    try
      let i' = M.find a ai in
      M.add a (fradd i i') ai
    with Not_found -> M.add a i ai in
  M.fold f ai ai'

(* reverse the graph *)
let rec reverse m =
  let rec add_from x yai n =
    M.fold (fun y ai n -> addes y x ai n) yai n in
  M.fold add_from m M.empty

(* collect the nodes of certain depth *)
let rec collectd m l n =
  if n <= 1 then l else
  let rec f l x =
    try
      M.fold (fun y _ l -> y::l) (M.find x m) l
    with Not_found -> l in
  collectd m (List.fold_left f [] l) (n-1)
  
(* split backward *)
let rec splitbwd count (ci,m,cv,qs) x =
  let rec flatten yai =
    M.fold (fun y -> M.fold (fun a i l -> (y,a,i)::l)) yai [] in
  let rec h x l xai =
    try
      let ai = M.find x xai in
      List.fold_left (fun xai x' -> M.add x' ai xai) (M.remove x xai) l 
    with Not_found -> xai in
  let v = if M.mem x cv then 1 else 0 in
  if not (M.mem x m) then (ci,m,cv,S.add x qs) else
  let k = flatten (M.find x m) in
  if List.length k + v <= 1 then (ci,m,cv,S.add x qs) else
  let l = List.map (fun (y,a,i) -> (newvar count x,y,a,i)) k in
  let l' = List.map (fun (x,_,_,_) -> x) l in
  let qs = List.fold_left (fun qs x -> S.add x qs) qs l' in
  let m = List.fold_left (fun m (x,y,a,i) -> M.add x (M.singleton y (M.singleton a i)) m) (M.remove x m) l in
  let (m,cv,l') =
    if v=1 then
      let i = M.find x cv in
      let xv = newvar count x in
      (M.add xv M.empty m, M.add xv i (M.remove x cv), xv::l')
    else (m,cv,l') in
  let ci =
    try
      let i = M.find x ci in
      List.fold_left (fun ci x' -> M.add x' i ci) (M.remove x ci) l'
    with Not_found -> ci in
  let m = M.map (h x l') m in
  (ci, m, cv, qs)

(* merge backward *)
let rec mergebwd ci m cv s =
  let rec f l n =
    match l with
    | [] -> raise Not_found
    | (x,yai)::ls ->
	if M.cardinal yai != 1 then f ls n else
	let (y,ai) = M.choose yai in
	if M.cardinal ai != 1 then f ls n else
	let (a,i) = M.choose ai in
	try
	  let (x',i') = M'.find (y,a) n in
	  ((x,i),(x',i'),y,a)
	with Not_found -> f ls (M'.add (y,a) (x,i) n) in
  let rec g x x' hi yai =
    let ai  = try M.find x  yai with Not_found -> M.empty in
    let ai' = try M.find x' yai with Not_found -> M.empty in
    if M.is_empty ai && M.is_empty ai' then yai
    else
      let ai' = M.map (fun j -> frmul j hi) ai' in 
      M.add x (mergee ai ai') (M.remove x' yai) in
  try
    let m' = M.fold (fun x i m -> adde x "" "" i m) cv m in
    let ((x,ij),(x',ij'),y,a) = f (List.filter (fun (x,_) -> S.mem x s) (M.bindings m')) M'.empty in
    let hi = frdiv ij' ij in
    let ci =
      let (i ,j ) = try M.find x  ci with Not_found -> (0,1) in
      let (i',j') = try M.find x' ci with Not_found -> (0,1) in
      if i=0 && i'=0 then ci
      else M.add x (fradd (i,j) (frmul (i',j') hi)) (M.remove x' ci) in
    let (m,cv) =
      if y="" then (M.remove x' m, M.add x ij (M.remove x' cv))
      else (M.add x (M.singleton y (M.singleton a ij)) (M.remove x' m), M.remove x' cv) in
    Some (ci, M.map (g x x' hi) m, cv)
  with Not_found -> None
let rec mergebwdd ci m cv s =
  match mergebwd ci m cv s with
  | Some (ci,m,cv) -> mergebwdd ci m cv s
  | None -> (ci,m,cv)

(* apply splitbwd and merge backward many times *)
let rec sbn ci m cv iter count =
  if count > iter then (ci,m,cv) else
  let l = collectd (reverse m) (M.fold (fun x _ l -> x::l) cv []) count in
  if l=[] then
    sbn ci m cv iter (count+1)
  else
    let (ci,m,cv,qs) = List.fold_left (splitbwd count) (ci,m,cv,S.empty) l in
    let (ci,m,cv) = if !merge then mergebwdd ci m cv qs else (ci,m,cv) in
    sbn ci m cv iter (count+1)
  
(* read the input file *)
let rec read_file f =
  let rec read_file3 f cv =
    try
      let s = input_line f in
      let ss = Str.split (Str.regexp "[ \t]+") s in
      match ss with
      | a::b::c::_ -> read_file3 f (M.add a (int_of_string b,int_of_string c) cv)
      | _ -> cv
    with End_of_file -> cv in
  let rec read_file2 f m =
    try
      let s = input_line f in
      let ss = Str.split (Str.regexp "[ \t]+") s in
      match ss with
      | a::b::c::d::e::_ -> read_file2 f (adde a b c (int_of_string d,int_of_string e) m)
      | _ -> let cv = read_file3 f M.empty in (m,cv)
    with End_of_file -> (m,M.empty) in
  let rec read_file1 f ci =
    try
      let s = input_line f in
      let ss = Str.split (Str.regexp "[ \t]+") s in
      match ss with
      | a::b::c::_ -> read_file1 f (M.add a (int_of_string b,int_of_string c) ci)
      | _ -> let (m,cv) = read_file2 f M.empty in (ci,m,cv)
    with End_of_file -> (ci,M.empty,M.empty) in
  read_file1 f M.empty

(* output file *)
let rec output_file f ci m cv =
  M.iter (fun x ij -> let (i,j) = mm ij in output_line f (String.concat " " [x;string_of_int i;string_of_int j])) ci;
  output_char f '\n';
  M.iter (fun x yai -> M.iter (fun y ai -> M.iter (fun a ij -> let (i,j) = mm ij in output_line f (String.concat " " [x;y;a;string_of_int i;string_of_int j])) ai) yai) m;
  output_char f '\n';
  M.iter (fun x ij -> let (i,j) = mm ij in output_line f (String.concat " " [x;string_of_int i;string_of_int j])) cv
    
(* main *)
let _ =
  let rec options i =
    if i >= Array.length Sys.argv then () else
    if Sys.argv.(i) = "-o" then (ovwr := true; options (i+1)) else
    if Sys.argv.(i) = "-nonmerge" then (merge := false; options (i+1)) in
  let fn = Sys.argv.(1) in
  let ifs = open_in fn in
  let (ci,m,cv) = read_file ifs in
  let itmin = max 1 (int_of_string Sys.argv.(3)) in  
  let itmax = min (M.cardinal m + M.cardinal cv) (int_of_string Sys.argv.(4)) in
  options 5;
  let ((ci,m,cv),s) =
    if Str.string_match (Str.regexp "-\\(fpe\\|sb\\|fwdl\\|fl\\|bwdr\\|br\\)") Sys.argv.(2) 0 then
      (sbn ci m cv itmax itmin,"fpe")
    else if Str.string_match (Str.regexp "-\\(bpe\\|sf\\|fwdr\\|fr\\|bwdl\\|bl\\)") Sys.argv.(2) 0 then
      let (cv,m,ci) = sbn cv (reverse m) ci itmax itmin in
      ((ci,reverse m, cv),"bpe")
    else raise (Invalid_argument Sys.argv.(2)) in
  let fm =
    if !ovwr then fn
    else String.concat "" [List.hd (Str.split (Str.regexp "[.]") (List.hd (List.rev (Str.split (Str.regexp "[/]") fn))));"_";s;".out"] in
  let ofs = open_out fm in
  output_file ofs ci m cv;
  close_in ifs;
  close_out ofs;
