#include "pt_simulation.h"



// for glpk
int add_elem2(int p[], int x, int q[], int y, double r[], double z, int s){
  if (z!=0){
    p[s+1] = x;
    q[s+1] = y;
    r[s+1] = z;
    s++;
  }
  return (s);
}
int solve_by_glpk(int a, int n, int m, int ***c, int *cv, int *ci, int ***d, int *dv, int *di, double *x, int flags){
  cout << "Making lp structure for glpk..." << endl;
  glp_prob *lp;
  lp = glp_create_prob();

  if (flags & SUMUNDERONE) glp_add_rows(lp,n+a*n*m+2*m);
  else glp_add_rows(lp,n+a*n*m+m);

  int i,j,k,l,h;

  // rhs & constr type
  if (!(flags & BWD)) {
    for (i=0; i<n; i++) glp_set_row_bnds(lp,i+1,GLP_LO,ci[i],0);
    for (i=0; i<a*m*n; i++) glp_set_row_bnds(lp,n+i+1,GLP_UP,0,0);
    for (i=0; i<m; i++) {
      if (dv[i] == 0) glp_set_row_bnds(lp,n+a*m*n+i+1,GLP_FX,0,0);
      else glp_set_row_bnds(lp,n+a*m*n+i+1,GLP_DB,0,dv[i]);
    }
  }
  else {
    for (i=0; i<n; i++) {
      if (ci[i] == 0) glp_set_row_bnds(lp,i+1,GLP_FX,0,0);
      else glp_set_row_bnds(lp,i+1,GLP_DB,0,ci[i]);
    }
    for (i=0; i<a*m*n; i++) glp_set_row_bnds(lp,n+i+1,GLP_LO,0,0);
    for (i=0; i<m; i++) glp_set_row_bnds(lp,n+a*m*n+i+1,GLP_LO,dv[i],0);
  }
  if (flags & SUMUNDERONE) for (i=0; i<m; i++) glp_set_row_bnds(lp,n+m+a*m*n+i+1,GLP_DB,0,1);

  glp_add_cols(lp,m*n);
  for (i=0; i<n*m; i++) glp_set_col_bnds(lp,i+1,GLP_LO,0,0);

  int *p, *q; double *r;
  int nz1 = 0, nz2 = 0;
  for (k=0; k<a; k++) for (i=0; i<n; i++) for (j=0; j<n; j++) if (c[k][i][j]!=0) nz1++;
  for (i=0; i<n; i++) if (cv[i]!=0) nz1++;
  if (flags & SUMUNDERONE) nz1+=n;
  for (k=0; k<a; k++) for (i=0; i<m; i++) for (j=0; j<m; j++) if (d[k][j][i]!=0) nz2++;
  for (i=0; i<m; i++) if (di[i]!=0) nz2++;
  p = new int [nz1*m+nz2*n+1];
  q = new int [nz1*m+nz2*n+1];
  r = new double [nz1*m+nz2*n+1];

  // make A
  int s = 0;
  for (i=0; i<n; i++)
    for (j=0; j<m; j++) s = add_elem2(p,i+1,q,i+j*n+1,r,di[j],s);  
  for (k=0; k<a; k++){
    for (l=0; l<m; l++){
      for (h=0; h<n; h++){
	for (i=0; i<m; i++){
	  if (l==i){
	    for (j=0; j<n; j++){
	      if (h==j) s = add_elem2(p,n+m*n*k+n*l+h+1,q,i*n+j+1,r,c[k][h][j]-d[k][l][i],s);
	      else      s = add_elem2(p,n+m*n*k+n*l+h+1,q,i*n+j+1,r,c[k][h][j],s);
	    }
	  }
	  else s = add_elem2(p,n+m*n*k+n*l+h+1,q,i*n+h+1,r,-d[k][l][i],s);
	}
      }
    }
  }
  for (i=0; i<m; i++)
    for (j=0; j<n; j++) s = add_elem2(p,n+m*n*a+i+1,q,n*i+j+1,r,cv[j],s);
  if (flags & SUMUNDERONE)
    for (i=0; i<m; i++)
      for (j=0; j<n; j++) s = add_elem2(p,n+m*n*a+m+i+1,q,n*i+j+1,r,1,s);
  cout << "Matrix ready..." << endl;

  // load matrix
  glp_load_matrix(lp,s,p,q,r);

  // solve
  int ret = 0;
  if (flags & INTERIOR){
    ret=glp_interior(lp,NULL);
    if (ret == 0) ret = 1;
    else ret = 0;
    if (ret && (flags & OUTPUT))
      for (i=0; i<n*m; i++) x[i] = glp_ipt_col_prim(lp,i+1);
  }
  else {
    // set parmeters
    glp_smcp parm;
    glp_init_smcp(&parm);
    parm.presolve = GLP_ON;
    ret = glp_simplex(lp,&parm);
    if (ret == 0) ret = 1;
    else ret = 0;
    if (ret && (flags & OUTPUT))
      for (i=0; i<n*m; i++) x[i] = glp_get_col_prim(lp,i+1);
  }
    

  glp_delete_prob(lp);

  return (ret);
}
      
